<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

use function strip_tags;
use function trim;

/**
 * Class Accordion
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class Accordion extends AbstractTabbed
{
    protected function getIdContext(): string
    {
        return 'accordion';
    }

    protected function canContentUserUseFaq(array $renderingOptions): bool
    {
        $entity = $renderingOptions['entity'] ?? null;
        if (($entity instanceof \XenAddons\AMS\Entity\ArticleItem) && $entity->hasPermission( 'svBbcodeAccordionFaqSeo'))
        {
            // AMS articles
            return true;
        }

        return false;
    }

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $tagOptions = $this->getTagOptions($tagOption, $options);
        $tagOptions = $this->parseTagOptions($tagOptions, $options);

        $containerId = $this->getUniqueContainerId($options);
        $slides = $this->getSlides($containerId, $tagChildren, $tagOptions, $tag, $options);
        if (!$slides)
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }
        $template = $this->rendererType === 'emailHtml' ? 'email:sv_bbcode_tabbed_content' : 'public:sv_bbcode_accordion';

        if ($tagOptions['faq'] ?? false)
        {
            $pageParams = $this->templater->pageParams;
            $ldStructuredData = $pageParams['svLdJson'] ?? [];

            // https://developers.google.com/search/docs/appearance/structured-data/faqpage
            // https://schema.org/FAQPage
            // https://schema.org/Question
            $ldStructuredData['@type'] = 'FAQPage';
            if (!isset($ldStructuredData['mainEntity']))
            {
                $ldStructuredData['mainEntity'] = [];
            }

            foreach ($slides as $slide)
            {
                $title = trim($slide['options']['title'] ?? '');
                $content = trim(strip_tags($slide['content'] ?? ''));
                if ($title === '' || $content === '')
                {
                    continue;
                }
                // html is allowed
                $content = trim($slide['content'] ?? '');

                $ldStructuredData['mainEntity'][] = [
                    'name' => $title,
                    'answerCount' => 1,
                    'acceptedAnswer' => [
                        'text' => $content,
                    ],
                ];
            }

            $this->templater->setPageParam('svLdJson', $ldStructuredData);
        }

        return $this->templater->renderTemplate($template, [
            'options'     => $tagOptions,
            'slides'      => $slides,
            'containerId' => $containerId,
        ]);
    }

    /**
     * @param string                $id
     * @param string|\XF\PreEscaped $title
     * @param string[]              $slideOptions
     * @param array                 $renderingOptions
     * @param string[]              $tagOptions
     * @return array
     */
    function parseSlideOptions(string $id, $title, array $slideOptions, array $renderingOptions, array $tagOptions): array
    {
        $constraints = $this->getOptionConstraints();
        $minHeight = $constraints['min_height'] ?? 22;

        $options = $this->getDefaultSlideOptions($tagOptions, $title);

        foreach ($slideOptions as $key => $slideOption)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'title':
                        $options['title'] = $slideOption;
                        break;
                    case 'open':
                        $options['open'] = $this->stringyTruth($slideOption);
                        break;
                    case 'align':
                        $this->parseTitleAlignmentOptions(\strtolower($slideOption), $options);
                        break;
                    case 'vertical_align':
                        $this->parseTitleAlignmentOptions(\strtolower($slideOption), $options, true);
                        break;
                    case 'height':
                        $options['height'] = $this->stringyFloat($slideOption);
                        break;
                }

                continue;
            }

            $original = $slideOption;
            $slideOption = \strtolower(\trim($slideOption));

            switch ($slideOption)
            {
                case 'left':
                case 'center':
                case 'right':
                    $options['title_align'] = $slideOption;
                    continue 2;

                //case 'top':
                //case 'bottom':
                //    $options['title_align_vertical'] = $slideOption;
                //    continue 2;

                case 'open':
                    $options['open'] = true;
                    continue 2;
            }

            if (\preg_match('#^\d+$#', $slideOption))
            {
                $options['height'] = $slideOption;
            }

            // title
            if ($slideOption !== '')
            {
                $options['title'] = $original;
            }
        }

        if ($options['height'] !== null)
        {
            switch($options['height_type'])
            {
                case 'px':
                    if ($options['height'] < $minHeight)
                    {
                        $options['height'] = $minHeight; // Min-height must be 22px to make overflow scroller visible
                    }
                    else if (!empty($constraints['max_height']) && $options['height'] > $constraints['max_height'])
                    {
                        $options['height'] = $constraints['max_height'];
                    }
                    break;
            }
        }

        return $options;
    }

    /**
     * @param string[]              $tagOptions
     * @param string|\XF\PreEscaped $title
     * @return array
     */
    protected function getDefaultSlideOptions(array $tagOptions, $title): array
    {
        $globalHeight = $tagOptions['height'] ?? null;
        $globalHeightType = $tagOptions['height_type'] ?? 'px';
        $defaultAlign = $tagOptions['title_align'];

        return [
            'title'                => $title,
            'title_align'          => $defaultAlign,
            'title_align_vertical' => '',
            'open'                 => false,
            'height'               => $globalHeight,
            'height_type'          => $globalHeightType,
        ];
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
        // unique options
        foreach ($lowerTagOptions as $key => $tagOption)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'align':
                    case 'title_align':
                        $this->parseTitleAlignmentOptions(\strtolower($tagOption), $options);
                        break;
                    case 'faq':
                        if ($this->canContentUserUseFaq($renderingOptions))
                        {
                            $options['faq'] = true;
                        }
                        break;
                }
                continue;
            }

            $this->parseTitleAlignmentOptions($tagOption, $options);
        }
    }

    public function getDefaultOptions(): array
    {
        $style = $this->getStyle();

        return [
            'width'       => $style->getProperty('svBbcodeAccordionDefaultWidth'),
            'width_type'  => $style->getProperty('svBbcodeAccordionDefaultWidthUnit'),
            'block_align' => $style->getProperty('svBbcodeAccordionDefaultBlockAlignment'),
            'title_align' => $style->getProperty('svBbcodeAccordionDefaultTitleAlignment'),
            'faq'         => (bool)$style->getProperty('svBbcodeAccordionFaqSeo'),
        ];
    }

    protected function getOptionConstraints(): array
    {
        return \array_merge(parent::getOptionConstraints(), [
            'allowHeight' => true,
            'max_width'   => $this->getStyleProperty('svBbcodeAccordionMaxWidth', 'svBbcodeAccordionApplyWidthRestriction'),
            'max_height'  => $this->getStyleProperty('svBbcodeAccordionMaxHeight', 'svBbcodeAccordionApplyHeightRestriction'),
            'min_height'  => $this->getStyleProperty('svBbcodeAccordionMinHeight'),
        ]);
    }
}