<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

use SV\AdvancedBbCodesPack\SimilarContent\SimilarContentContainer;
use SV\AdvancedBbCodesPack\SimilarContent\SimilarContentKeywords;
use SV\ElasticSearchEssentials\Search\Data\RestrictionType;
use SV\ElasticSearchEssentials\Service\SimilarContentFinder;
use SV\SearchImprovements\XF\Repository\Search as SearchRepo;
use SV\StandardLib\Helper;
use XF\Http\Request;
use XF\Mvc\Entity\AbstractCollection;
use XF\Mvc\Entity\ArrayCollection;
use XF\Mvc\Entity\Entity;
use XF\Search\Data\AbstractData;
use XF\Search\Query\Query;
use XF\Search\RenderWrapper;
use function array_filter;
use function array_key_exists;
use function array_merge;
use function assert;
use function count;
use function in_array;
use function json_decode;
use function json_last_error;
use function strtolower;
use function trim;

/**
 * A BB code renderer for [search_content] tags.
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class SearchContent extends AbstractTag
{
    /**
     * @param mixed $entity
     * @return bool
     */
    protected function isArticleManagementSystemItem($entity): bool
    {
        return ($entity instanceof \XenAddons\AMS\Entity\ArticleItem) || ($entity instanceof \XenAddons\AMS\Entity\ArticlePage);
    }

    protected function canRender(array $options): bool
    {
        // todo: handler post:preview context
        $entity = $options['entity'] ?? null;
        if (\XF::options()->svRestrictSimilarContentBbCode ?? false)
        {
            if ($entity instanceof \XF\Entity\Post && $entity->isFirstPost())
            {
                $thread = $entity->Thread;
                if ($thread !== null && $thread->discussion_type === 'article')
                {
                    return true;
                }
            }
            if ($this->isArticleManagementSystemItem($entity))
            {
                return true;
            }
            else if ($entity instanceof \XF\Entity\Page)
            {
                return true;
            }
            else if ($this->renderer->getRules()->getSubContext() === 'preview')
            {
                return true;
            }

            return false;
        }

        return true;
    }

    protected $genericCriteria = [
        'title',
        'title_only',
        'keywords',
        'type',
        'content',
        'similar',
        'limit',
        'newer_than_update',
        'newer_than',
        'older_than',
        'order',
    ];

    /**
     * @noinspection PhpUnnecessaryLocalVariableInspection
     * @noinspection PhpUnusedParameterInspection
     */
    protected function getCriteria(array $tagChildren, array $filteredTagOptions, array $tagOptions, array $options): array
    {
        if ($tagOptions['json'] ?? false)
        {
            $json = $this->renderer->renderSubTreePlain($tagChildren);
            $criteria = @json_decode($json, true) ?: [];
        }
        else
        {
            $criteria = $filteredTagOptions;
            foreach ($criteria as $key => $value)
            {
                if (in_array($key, $this->genericCriteria, true))
                {
                    continue;
                }
                $blob = @json_decode($value, true);
                if (json_last_error() === \JSON_ERROR_NONE)
                {
                    $criteria[$key] = $blob;
                }
            }

            if (!array_key_exists('title', $criteria))
            {
                $criteria['title'] = trim($this->renderer->renderSubTreePlain($tagChildren));
            }
        }
        $criteria['keywords'] = $criteria['keywords'] ?? $tagOptions['keywords'] ?? '*';
        if ($criteria['keywords'] === '')
        {
            $criteria['keywords'] = '*';
        }
        $criteria['type'] = $criteria['type'] ?? $tagOptions['type'] ?? 'thread';
        $criteria['similar'] = $this->stringyTruth($criteria['similar'] ?? $tagOptions['similar'] ?? true);
        $criteria['title_only'] = $this->stringyTruth($criteria['title_only'] ?? $tagOptions['title_only'] ?? true);
        $criteria['limit'] = $this->stringyInt($criteria['limit'] ?? $tagOptions['limit'] ?? 10);
        $criteria = array_filter($criteria, function ($v) {
            return $v !== null && $v !== '' && $v !== [];
        });

        return $criteria;
    }

    protected function getSearchDataFromCriteria(array $criteria): Request
    {
        $searchData = [
            'search_type' => $criteria['type'] ?? 'thread',
            'keywords'    => $criteria['keywords'] ?? '*',
            'order'       => $criteria['order'] ?? '',
            'c'           => $criteria,
        ];
        foreach ($this->genericCriteria as $key)
        {
            unset($searchData['c'][$key]);
        }

        return new Request(\XF::app()->inputFilterer(), $searchData, [], [], []);
    }

    protected function applyGenericCriteria(Query $query, SimilarContentFinder $similarContentFinder, array $criteria): void
    {
        $value = (int)($criteria['limit'] ?? 0);
        if ($value !== 0)
        {
            $similarContentFinder->setLimit($value);
        }

        $value = (int)($criteria['newer_than_update'] ?? 0);
        if ($value !== 0)
        {
            $similarContentFinder->setLastUpdatedDateLimit($value);
        }

        $value = (int)($criteria['newer_than'] ?? 0);
        if ($value !== 0)
        {
            $similarContentFinder->setDateLimit($value);
        }

        $value = (int)($criteria['older_than'] ?? 0);
        if ($value !== 0)
        {
            $query->olderThan($value);
        }

        $value = (bool)($criteria['title_only'] ?? false);
        if ($value)
        {
            $similarContentFinder->getQuery()->inTitleOnly();
        }

        $value = $criteria['order'] ?? '';
        if ($value !== '')
        {
            $similarContentFinder->setSort($value);
            $query->orderedBy($similarContentFinder->getSort());
        }
    }

    protected function rewriteQueryToChild(string $contentType, AbstractData $handler): AbstractData
    {
        $app = \XF::app();
        $searchRepo = $app->repository('XF:Search');
        assert($searchRepo instanceof SearchRepo);
        $firstChildType = $searchRepo->getChildContentTypeForContainerType($contentType);
        if ($firstChildType !== null && $firstChildType !== $contentType)
        {
            $handler = $app->search()->handler($firstChildType);
            assert($handler !== null);

            return $handler;
        }

        return $handler;
    }

    protected function getSimilarContentFinder(SimilarContentContainer $container, SimilarContentKeywords $content): SimilarContentFinder
    {
        /** @var SimilarContentFinder $similarContentFinder */
        $similarContentFinder = $this->app->service(SimilarContentFinder::class, $container, $content);
        $options = \XF::options();

        $cacheTime = \XF::$developmentMode
            ? 0
            : (int)($options->svSearchContentBbCodeCacheTime ?? $options->svESE_SimilarContentOnCreateCache ?? 300);
        $similarContentFinder->setCacheTime($cacheTime);
        $similarContentFinder->setSort($content->criteria['order'] ?? 'relevance');
        $similarContentFinder->setLimit((int)($content->criteria['limit'] ?? 25));
        // todo: implement XF options for this bb-code
        //$similarContentFinder->setDateLimit($options->svESE_SimilarContentOnCreateDateLimit ?? 0);
        //$similarContentFinder->setLastUpdatedDateLimit($options->svESE_AutocompleteLastUpdatedLimit ?? 0);
        $similarContentFinder->setCutOffFrequency($options->svESE_SimilarContentOnCreateCutoffFreq ?? 0);

        $container->similarContentFinder = $similarContentFinder;
        $container->content = $content;
        $content->onPrepare = function(SimilarContentKeywords $content) use ($similarContentFinder) {
            $this->prepareSimilarContentService($content, $similarContentFinder);
        };

        return $similarContentFinder;
    }

    protected function setupSimilarContentFinder(SimilarContentFinder $similarContentFinder, Request $request, array $tagOptions): SimilarContentFinder
    {
        $autoCompleter = $similarContentFinder->getAutoCompleter();
        $autoCompleter->setInputRequest($request);
        $autoCompleter->setMinimumQueryLength(0);
        $autoCompleter->setApplyingConstraints(false);
        if (!($tagOptions['similar'] ?? true))
        {
            $autoCompleter->setSearchEssTitle(false);
        }

        return $similarContentFinder;
    }

    protected function prepareSimilarContentService(SimilarContentKeywords $similarContentKeywords, SimilarContentFinder $similarContentFinder): void
    {
        $tagOptions = $similarContentKeywords->options;

        $autoCompleter = $similarContentFinder->getAutoCompleter();
        $query = $autoCompleter->getQuery();
        $query->setAutoCompleteRestrictionType(RestrictionType::SIMILAR_CONTENTS);
        if (!($tagOptions['similar'] ?? true))
        {
            // undo various prepare bits
            $query->setSupportQuoteForExact(true)
                  ->setMatchQuery(false)
                  ->setMultiMatchQuery(false)
            ;
            $query->withKeywords($similarContentKeywords->criteria['keywords'], $query->getTitleOnly());
        }

        $query = $similarContentFinder->getQuery();
        $this->applyGenericCriteria($query, $similarContentFinder, $similarContentKeywords->criteria);
        // see SimilarContentKeywords::applySimilarContentMetadataToQuery for non-generic constraints
    }

    public function getDefaultOptions(): array
    {
        $style = $this->getStyle();

        return [
            'similar' => true,
            'title_only' => false,
            'title'   => '',
            'limit'   => 10,

            'preview'     => (bool)$style->getProperty('svBbcodeSearchDefaultPreview'),
            'style'       => $style->getProperty('svBbcodeSearchDefaultStyle'),
            'block_align' => $style->getProperty('svBbcodeSearchDefaultBlockAlignment'),
            'title_align' => $style->getProperty('svBbcodeSearchDefaultTitleAlignment'),
            'width'       => $style->getProperty('svBbcodeSearchDefaultWidth'),
            'width_type'  => $style->getProperty('svBbcodeSearchDefaultWidthUnit'),
            'height'      => null,
            'height_type' => null,
        ];
    }

    protected function getOptionConstraints(): array
    {
        return array_merge(parent::getOptionConstraints(), [
            'allowHeight' => true,
        ]);
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
        foreach ($lowerTagOptions as $key => $tagOption)
        {
            if (is_integer($key))
            {
                continue;
            }

            switch ($key)
            {
                case 'json':
                case 'preview':
                case 'title_only':
                case 'similar':
                    $options[$key] = $this->stringyTruth($tagOption);
                    break;
                case 'title':
                    $options['title'] = $tagOption;
                    break;
                case 'limit':
                    $options['limit'] = (int)min($this->stringyInt($tagOption), \XF::options()->maximumSearchResults);
                    break;
                case 'style':
                    $options['style'] = $tagOption;
                    break;
                case 'title_align':
                    $this->parseTitleAlignmentOptions(strtolower($tagOption), $options, false);
                    break;
            }
        }
    }

    /**
     * @param array                $tagChildren
     * @param null|string|string[] $tagOption
     * @param array                $tag
     * @param array                $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        if (!$this->canRender($options))
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }

        if (!Helper::isAddOnActive('SV/ElasticSearchEssentials', '3.15.4'))
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }

        $rawTagOptions = $this->getTagOptions($tagOption, $options);
        if (count($rawTagOptions) === 0)
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }
        $tagOptions = $this->parseTagOptions($rawTagOptions, $options);
        $filteredTagOptions = $rawTagOptions;
        if (count($filteredTagOptions) !== 0)
        {
            foreach ($tagOptions as $key => $value)
            {
                unset($filteredTagOptions[$key]);
            }
        }
        $criteria = $this->getCriteria($tagChildren, $filteredTagOptions, $tagOptions, $options);
        $tagOptions = array_merge($tagOptions, $criteria);
        unset($criteria['title']);

        $title = $tagOptions['title'] ?? '';
        $contentType = $criteria['type'];

        $searcher = $this->app->search();
        if ($contentType === '' || !$searcher->isValidContentType($contentType))
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }
        $handler = $searcher->handler($contentType);
        $handler = $this->rewriteQueryToChild($contentType, $handler);
        /** @var \SV\ElasticSearchEssentials\XF\Repository\Search $searchRepo */
        $searchRepo = $this->app->repository('XF:Search');
        $searchHandlers = $searchRepo->getSearchHandlers();

        $request = $this->getSearchDataFromCriteria($criteria);
        $container = SimilarContentContainer::factory($handler, $contentType, $criteria, $request);
        $content = SimilarContentKeywords::factory($handler, $contentType, $criteria, $request, $tagOptions);

        $similarContentFinder = $this->getSimilarContentFinder($container, $content);
        $similarContentFinder = $this->setupSimilarContentFinder($similarContentFinder, $request, $tagOptions);

        /** @var AbstractCollection $similarContents */
        $similarContents = $similarContentFinder->getSimilarContents() ?? new ArrayCollection([]);

        $results = [];
        foreach ($similarContents as $result)
        {
            /** @var Entity $result */
            $handler = $searchHandlers[$result->getEntityContentType()]['handler'] ?? null;
            if ($handler !== null)
            {
                $results[] = new RenderWrapper($handler, $result, []);
            }
        }

        $entity = $options['entity'] ?? null;
        $addBbWrapper = $this->isArticleManagementSystemItem($entity);

        return $this->templater->renderTemplate('public:sv_bbcode_search', [
            'title'           => $title,
            'similarContents' => $results,
            'criteria'        => $criteria,
            'options'         => $tagOptions,
            'addBbWrapper'    => $addBbWrapper,
        ]);
    }
}