<?php

namespace SV\AdvancedBbCodesPack;

use SV\StandardLib\InstallerHelper;
use XF\AddOn\AbstractSetup;
use XF\AddOn\StepRunnerInstallTrait;
use XF\AddOn\StepRunnerUninstallTrait;
use XF\AddOn\StepRunnerUpgradeTrait;
use function array_fill_keys;
use function array_key_exists;
use function array_keys;
use function file_exists;
use function in_array;

/**
 * Class Setup
 *
 * @package SV\AdvancedBbCodesPack
 */
class Setup extends AbstractSetup
{
    use InstallerHelper;
    use StepRunnerInstallTrait;
    use StepRunnerUpgradeTrait;
    use StepRunnerUninstallTrait;

    public function installStep1(): void
    {
        $sm = $this->schemaManager();

        foreach ($this->getTables() as $tableName => $callback)
        {
            $sm->createTable($tableName, $callback);
            $sm->alterTable($tableName, $callback);
        }
    }

    public function installStep2(): void
    {
        $sm = $this->schemaManager();

        foreach ($this->getAlterTables() as $tableName => $callback)
        {
            if ($sm->tableExists($tableName))
            {
                $sm->alterTable($tableName, $callback);
            }
        }
    }

    public function installStep3(): void
    {
        $this->migrateOptions();
        $this->initializeOptionToStylePropertyConversion(
            'sv_bbcode_opt2style',
            $this->getOptionToStylePropertyConversionMap()
        );
    }

    public function installStep4(): void
    {
        $this->importMissingBbCodeMulti([
            'hr',
            'h1',
            'h2',
            'h3',
            'h4',
        ], $this->getBbCodeFiles());
    }

    public function upgrade1000011Step1(): void
    {
        $this->migrateOptions();
        $this->initializeOptionToStylePropertyConversion(
            'sv_bbcode_opt2style',
            $this->getOptionToStylePropertyConversionMap()
        );
    }

    public function upgrade1010100Step1(): void
    {
        $this->importMissingBbCode([
            'article',
            'bimg',
            'encadre',
            'fieldset',
            'hr',
            'justify',
            'latex',
            'slider',
            'xtable',
        ]);
    }

    public function upgrade1010200Step1(): void
    {
        $this->importMissingBbCode([
            'bimg',
        ]);
    }

    public function upgrade1010402Step1(): void
    {
        $this->importMissingBbCode([
            'accordion',
            'tabs',
            'slider',
        ]);
    }

    public function upgrade1020101Step1(): void
    {
        $this->importMissingBbCode([
            'hr',
            'bimg',
        ]);
    }

    public function upgrade1020000Step1(): void
    {
        $this->importMissingBbCode([
        ]);
    }

    public function upgrade1070500Step1(): void
    {
        $this->renameStyleProperty('svBbCodeEditorWordCoutTimeOut', 'svBbCodeEditorWordCountTimeOut');
    }

    public function upgrade1071301Step1(): void
    {
        $this->importMissingBbCode([
            'accordion',
            'bimg',
            'encadre',
            'fieldset',
            'slider',
            'tabs',
        ]);
    }

    public function upgrade1090003Step1(): void
    {
        $this->importMissingBbCodeMulti([
            'hr',
            'h1',
            'h2',
            'h3',
            'h4',
        ], ['bb_codes.xml' => true, 'headers.xml' => true]);
    }

    public function upgrade1110400Step1(): void
    {
        /** @var \XF\Entity\Option $option */
        $option = \XF::app()->find('XF:Option', 'svMathJaxSource');
        if ($option !== null && $option->option_value === 'jsdelivr')
        {
            $option->option_value = 'cloudflare';
            $option->save();
        }
    }

    public function upgrade1685001479Step1(): void
    {
        $this->importMissingBbCodeMulti([
        ], $this->getBbCodeFiles());
    }

    public function upgrade1685001479Step2(): void
    {
        $this->removeBbCode('justify');
    }

    public function upgrade1691014457Step1(): void
    {
        $this->importMissingBbCodeMulti([
            'searchcontent'
        ], ['search_content.xml' => false]);
    }

    public function upgrade1698628234Step1(): void
    {
        $this->importMissingBbCodeMulti([
        ], ['user_group.xml' => false]);
    }

    public function uninstallStep1(): void
    {
        $this->removeAddOnBbCodes();
    }

    public function uninstallStep2(): void
    {
        $sm = $this->schemaManager();

        foreach ($this->getTables() as $tableName => $callback)
        {
            $sm->dropTable($tableName);
        }
    }

    public function uninstallStep3(): void
    {
        $sm = $this->schemaManager();

        foreach ($this->getRemoveAlterTables() as $tableName => $callback)
        {
            $sm->alterTable($tableName, $callback);
        }
    }

    public function uninstallStep4(): void
    {
        $this->deletePhrases(['svAdvancedBbCode_word_count.%']);
    }

    public function postInstall(array &$stateChanges): void
    {
        $this->postUpgrade(0, $stateChanges);
    }

    /**
     * @param int   $previousVersion
     * @param array $stateChanges
     */
    public function postUpgrade($previousVersion, array &$stateChanges): void
    {
        $previousVersion = (int)$previousVersion;
        if ($previousVersion < 1000006)
        {
            $this->finalizeOptionToStylePropertyConversion(
                'sv_bbcode_opt2style'
            );
        }
        if ($previousVersion < 1697916422)
        {
            /** @var \XF\Repository\Option $optionRepo */
            $optionRepo = \XF::repository('XF:Option');
            $optionRepo->updateOption('svNormalizeQuoteCharactersOnRender', \XF::options()->svNormalizeQuoteCharactersOnPaste);
        }
        $this->importMissingBbCode();
        $this->deownBbCode();
    }

    /**
     * Migrate options from other add-ons.
     */
    protected function migrateOptions(): void
    {
        $this->ownOption('modInterruptTags');
        $this->ownOption('modInterruptThreadmarkCategory');

        $this->renamePhrases([
            'Sedo_AdvBBcodeBar_text_gview' => 'sv_bbcode.text_gview',
        ]);
    }

    protected function ownOption(string $name): void
    {
        /** @var \XF\Entity\Option $option */
        $option = \XF::finder('XF:Option')->whereId($name)->fetchOne();
        if ($option)
        {
            $option->addon_id = $this->addOn->getAddOnId();
            $option->saveIfChanged();
        }
    }

    protected function initializeOptionToStylePropertyConversion(
        string $registryKey,
        array  $conversionMap
    ): void
    {
        $values = [];
        $options = $this->app->options();
        foreach ($conversionMap as $optionId => $propertyName)
        {
            $value = $options->{$optionId} ?? null;
            if ($value === null)
            {
                continue;
            }

            $values[$propertyName] = $value;
        }

        $this->app->registry()->set($registryKey, $values);
    }

    protected function finalizeOptionToStylePropertyConversion(string $registryKey): void
    {
        if (!$registryKey)
        {
            return;
        }

        /** @var \XF\Repository\Style $styleRepo */
        $styleRepo = $this->app->repository('XF:Style');
        $styles = $styleRepo->findStyles()
                            ->where('parent_id', 0)
                            ->fetch();

        $values = $this->app->registry()->get($registryKey);
        if (!$values)
        {
            return;
        }

        $db = $this->db();
        $db->beginTransaction();

        /** @var \XF\Repository\StyleProperty $stylePropertyRepo */
        $stylePropertyRepo = $this->app->repository('XF:StyleProperty');

        foreach ($styles as $style)
        {
            /** @var \XF\Entity\Style $style */
            $stylePropertyRepo->updatePropertyValues($style, $values);
        }

        $this->app->registry()->delete($registryKey);
        $db->commit();
    }

    protected function getOptionToStylePropertyConversionMap(): array
    {
        return [
            // bimg
            'AdvBBcodeBar_imgdefault'                        => 'svBbcodeBimgDefaultWidth',
            'AdvBBcodeBar_imgmax'                            => 'svBbcodeBimgMaxWidth',

            // gview
            'AdvBBcodeBar_gview_height'                      => 'svBbcodeGviewHeight',

            // encadre
            'AdvBBcodeBar_encdefault'                        => 'svBbcodeEncadreDefaultWidth',
            'AdvBBcodeBar_encdefault_defaultwidth_unit'      => 'svBbcodeEncadreDefaultWidthUnit',
            'AdvBBcodeBar_encmax_px'                         => 'svBbcodeEncadreMaxWidthPx',
            'AdvBBcodeBar_encmax'                            => 'svBbcodeEncadreMaxWidthPercent',

            // table
            'quattro_extra_bbcodes_xtable_minwidth_px'       => 'svBbcodeTableMinWidthPx',
            'quattro_extra_bbcodes_xtable_maxwidth_px'       => 'svBbcodeTableMaxWidthPx',
            'quattro_extra_bbcodes_xtable_minheight_px'      => 'svBbcodeTableMinHeightPx',
            'quattro_extra_bbcodes_xtable_maxheight_px'      => 'svBbcodeTableMaxHeightPx',
            'quattro_extra_bbcodes_xtable_minwidth_percent'  => 'svBbcodeTableMinWidthPercent',
            'quattro_extra_bbcodes_xtable_maxwidth_percent'  => 'svBbcodeTableMaxWidthPercent',
            'quattro_extra_bbcodes_xtable_minheight_percent' => 'svBbcodeTableMinHeightPercent',
            'quattro_extra_bbcodes_xtable_maxheight_percent' => 'svBbcodeTableMaxHeightPercent',
            'quattro_extra_bbcodes_xtable_cellpadding_max'   => 'svBbcodeTableMaxCellpadding',
            'quattro_extra_bbcodes_xtable_cellspacing_max'   => 'svBbcodeTableMaxCellspacing',
            'quattro_extra_bbcodes_xtable_border_max'        => 'svBbcodeTableMaxBorderWidth',

            // accordion
            'AdvBBcodeBar_accordion_defaultwidth'            => 'svBbcodeAccordionDefaultWidth',
            'AdvBBcodeBar_accordion_defaultwidth_unit'       => 'svBbcodeAccordionDefaultWidthUnit',
            'AdvBBcodeBar_accordion_maxwidth'                => 'svBbcodeAccordionMaxWidth',
            'AdvBBcodeBar_accordion_slides_maxheight'        => 'svBbcodeAccordionMaxHeight',
            'AdvBBcodeBar_accordion_titles_defaultalign'     => 'svBbcodeAccordionDefaultTitleAlignment',
            'AdvBBcodeBar_accordion_slides_easing_effect'    => 'svBbcodeAccordionEasingEffect',
            'AdvBBcodeBar_accordion_slides_easing_duration'  => 'svBbcodeAccordionEasingDuration',

            // tabs
            'AdvBBcodeBar_tabs_default_width'                => 'svBbcodeTabsDefaultWidth',
            'AdvBBcodeBar_tabs_defaultwidth_unit'            => 'svBbcodeTabsDefaultWidthUnit',
            'AdvBBcodeBar_tabs_defaultheight'                => 'svBbcodeTabsDefaultHeight',
            'AdvBBcodeBar_tabs_maxwidth'                     => 'svBbcodeTabsMaxWidth',
            'AdvBBcodeBar_tabs_maxheight'                    => 'svBbcodeTabsMaxHeight',

            // slider
            'AdvBBcodeBar_slider_defaultwidth'               => 'svBbcodeSliderDefaultWidth',
            'AdvBBcodeBar_slider_defaultwidth_unit'          => 'svBbcodeSliderDefaultWidthUnit',
            'AdvBBcodeBar_slider_defaultheight'              => 'svBbcodeSliderDefaultHeight',
            'AdvBBcodeBar_slider_maxwidth'                   => 'svBbcodeSliderMaxWidth',
            'AdvBBcodeBar_slider_maxheight'                  => 'svBbcodeSliderMaxHeight',
            'AdvBBcodeBar_slider_interval_default'           => 'svBbcodeSliderDefaultInterval',
            'AdvBBcodeBar_slider_interval_min'               => 'svBbcodeSliderMinInterval',
            'AdvBBcodeBar_slider_interval_max'               => 'svBbcodeSliderMaxInterval',
            'AdvBBcodeBar_slider_autoplay_authorise'         => 'svBbcodeSliderAllowAutoplay',

            // fieldset
            'AdvBBcodeBar_fieldsetdefault'                   => 'svBbcodeFieldsetDefaultWidth',
            'AdvBBcodeBar_fieldsetdefault_width_unit'        => 'svBbcodeFieldsetDefaultWidthUnit',
            'AdvBBcodeBar_fieldsetmax_px'                    => 'svBbcodeFieldsetMaxWidthPx',
            'AdvBBcodeBar_fieldsetmax'                       => 'svBbcodeFieldsetMaxWidthPercent',

            // latex
            'AdvBBcodeBar_latex_defaultwidth'                => 'svBbcodeLatexDefaultWidth',
            'AdvBBcodeBar_latex_defaultwidthunit'            => 'svBbcodeLatexDefaultWidthUnit',
            'AdvBBcodeBar_latex_maxwidth_px'                 => 'svBbcodeLatexMaxWidthPx',
            'AdvBBcodeBar_latex_maxwidth_percent'            => 'svBbcodeLatexMaxWidthPercent',
            'AdvBBcodeBar_latex_maxheight'                   => 'svBbcodeLatexMaxHeight'
        ];
    }

    /**
     * To allow users to edit add-on owned, bb-code we need to remove the owning add-on :(
     */
    protected function deownBbCode(): void
    {
        $db = \XF::db();
        $db->beginTransaction();
        /** @var \XF\Entity\BbCode[] $bbcodes */
        $bbcodes = \XF::finder('XF:BbCode')->where('addon_id', '=', 'SV/AdvancedBbCodesPack')->fetch();
        foreach ($bbcodes as $bbcode)
        {
            $bbcode->addon_id = '';
            $bbcode->save(true, false);
        }
        $db->commit();
    }

    protected function loadXmlFile(string $filename): ?\SimpleXMLElement
    {
        if (!file_exists($filename))
        {
            return null;
        }

        try
        {
            /** @var ?\SimpleXMLElement $xml */
            $xml = \XF\Util\Xml::openFile($filename, true) ?: null;
        }
        catch (\Exception $e)
        {
            $xml = null;
        }

        return $xml;
    }

    protected function loadBbCodeFromXML(string $name): ?\SimpleXMLElement
    {
        $addonDir = \XF::getAddOnDirectory();
        $xml = $this->loadXmlFile("{$addonDir}/SV/AdvancedBbCodesPack/bb-codes/{$name}");
        if ($xml === null)
        {
            $xml = $this->loadXmlFile("{$addonDir}/SV/AdvancedBbCodesPack/_no_upload/bb-codes/{$name}");
        }

        return $xml;
    }

    protected function getBbCodeIds(\SimpleXMLElement $xml, array $skipKeys = []): array
    {
        $xmlBbCodes = $xml->bb_code;
        $skipKeys = array_fill_keys($skipKeys, true);

        $bbCodeIds = [];
        foreach ($xmlBbCodes as $xmlBbCode)
        {
            $bbCodeId = (string)$xmlBbCode['bb_code_id'];
            if (isset($skipKeys[$bbCodeId]))
            {
                continue;
            }

            $bbCodeIds[] = $bbCodeId;
        }

        return $bbCodeIds;
    }

    protected function removeBbCode(string $bbCode): void
    {
        $entity = \XF::app()->find('XF:BbCode', $bbCode);
        if ($entity !== null)
        {
            $entity->delete();
        }
    }

    protected function removeAddOnBbCodes(): void
    {
        $db = \XF::db();
        $db->beginTransaction();
        /** @var \XF\Entity\BbCode[] $bbcodes */
        $bbcodes = \XF::finder('XF:BbCode')->whereOr([
            'callback_class', 'like', $db->escapeLike('SV\AdvancedBbCodesPack\BbCode') . '%'
        ], [
            'addon_id', '=', 'SV/AdvancedBbCodesPack'
        ])->fetch();
        foreach ($bbcodes as $bbcode)
        {
            $bbcode->delete(true, false);
        }
        $db->commit();
    }

    /**
     * @param array<string> $forceUpdate
     * @param array<string, bool> $files
     */
    protected function importMissingBbCodeMulti(array $forceUpdate = [], array $files = []): void
    {
        foreach ($files as $file => $defaultEnabled)
        {
            $this->importMissingBbCode($forceUpdate, $file, $defaultEnabled);
        }
    }

    /**
     * @param string[]    $forceUpdate
     * @param string|null $file
     */
    protected function importMissingBbCode(array $forceUpdate = [], string $file = null, bool $defaultEnabled = true): void
    {
        $xml = $this->loadBbCodeFromXML($file ?: 'bb_codes.xml');
        if ($xml === null)
        {
            return;
        }
        $bbCodeIds = $this->getBbCodeIds($xml, []);

        $xmlBbCodes = $xml->bb_code;
        /** @var \XF\Entity\BbCode[] $existingBbCodes */
        $existingBbCodes = \XF::finder('XF:BbCode')
                      ->where('bb_code_id', $bbCodeIds)
                      ->order('bb_code_id')
                      ->fetch();

        $wasEnabled = [];
        foreach ($existingBbCodes as $bbCode)
        {
            if ($bbCode->active)
            {
                $wasEnabled[$bbCode->bb_code_id] = true;
            }
            if (in_array($bbCode->bb_code_id, $forceUpdate, true))
            {
                continue;
            }
            // remove from the XML to avoid force updating
            foreach ($xmlBbCodes as $xmlBbCode)
            {
                $bbCodeId = (string)$xmlBbCode['bb_code_id'];
                if ($bbCodeId === $bbCode->bb_code_id)
                {
                    $node = dom_import_simplexml($xmlBbCode);
                    $node->parentNode->removeChild($node);
                    break;
                }
            }
        }

        if ($xmlBbCodes->count() === 0)
        {
            return;
        }

        // default enable policy; import but leave disabled if they do not exist
        $toDisable = [];
        if (!$defaultEnabled)
        {
            foreach ($xmlBbCodes as $xmlBbCode)
            {
                $bbCodeId = (string)$xmlBbCode['bb_code_id'];
                if (array_key_exists($bbCodeId, $wasEnabled))
                {
                    continue;
                }
                $toDisable[$bbCodeId] = true;
            }
        }

        try
        {
            $db = \XF::db();
            $db->beginTransaction();
            /** @var \XF\Service\BbCode\Import $importer */
            $importer = \XF::service(\XF\Service\BbCode\Import::class);
            $importer->import($xml);

            if ($toDisable)
            {
                /** @var \XF\Entity\BbCode[] $bbCodes */
                $bbCodes = \XF::finder('XF:BbCode')->whereId(array_keys($toDisable))->fetch();
                foreach ($bbCodes as $bbCode)
                {
                    $bbCode->active = false;
                    $bbCode->save(true, false);
                }
            }

            $db->commit();
        }
        catch (\Exception $e)
        {
            \XF::logException($e, true, 'Error importing:' . $file . ', ');
        }
    }

    protected function getTables(): array
    {
        return [];
    }

    protected function getAlterTables(): array
    {
        return [];
    }

    protected function getRemoveAlterTables(): array
    {
        return [];
    }

    protected function getBbCodeFiles(): array
    {
        return [
            'bb_codes.xml' => true,
            'headers.xml' => true,
            //'legacy.xml' => false,
            'font_awesome.xml' => false,
            'search_content.xml' => false,
            'user_group.xml' => false,
        ];
    }
}