<?php

namespace SV\AdvancedBbCodesPack\SimilarContent;

use SV\ElasticSearchEssentials\Entity\ISimilarContentContainer;
use SV\ElasticSearchEssentials\Search\IEncodeForSimilarContentCache;
use XF\Http\Request;
use XF\Mvc\Entity\AbstractCollection;
use XF\Mvc\Entity\ArrayCollection;
use XF\Mvc\Entity\Entity;
use XF\Search\Data\AbstractData;
use function assert;
use function md5;
use function method_exists;

class SimilarContentContainer implements ISimilarContentContainer, IEncodeForSimilarContentCache
{
    /** @var string */
    public $contentType;
    /** @var array */
    public $criteria;
    /**  @var AbstractData */
    public $handler;
    /** @var Request */
    public $request;
    /** @var SimilarContentKeywords|null */
    public $content;
    /** @var \SV\ElasticSearchEssentials\Service\SimilarContentFinder|null */
    public $similarContentFinder;

    protected function __construct(AbstractData $handler, string $contentType, array $criteria, Request $request)
    {
        $this->handler = $handler;
        $this->contentType = $contentType;
        $this->criteria = $criteria;
        $this->request = $request;
    }

    public static function factory(AbstractData $handler, string $contentType, array $criteria, Request $request): self
    {
        $class = static::class;
        $class = \XF::extendClass($class);
        return new $class($handler, $contentType, $criteria, $request);
    }

    public function canViewSimilarContentsOnCreation(&$error = null): bool
    {
        return true;
    }

    public function canViewSimilarContentsOnQuickCreation(&$error = null): bool
    {
        return true;
    }

    public function canUseAutoComplete(&$error = null): bool
    {
        return true;
    }

    public function getCacheIdForSimilarContents(string $keywords): string
    {
        $permissionCombinationId  = \XF::visitor()->permission_combination_id ?: 1;
        return 'ESESimilar_bbcode_' . md5(json_encode($this->criteria)) . '_' . $permissionCombinationId;
    }

    public function getContentsFromCacheForSimilarContents(array $existingCache, int $limit): AbstractCollection
    {
        $onPrepare = $this->content->onPrepare ?? null;
        if ($onPrepare !== null)
        {
            $onPrepare($this->content);
        }
        $limit = $this->similarContentFinder->getLimit() ?? $limit;

        /** @var \SV\ElasticSearchEssentials\XF\Repository\Search $searchRepo */
        $searchRepo = \XF::app()->repository('XF:Search');
        $searchHandlers = $searchRepo->getSearchHandlers();

        $contentsByType = [];
        $sortKey = [];
        foreach ($existingCache as $contentTypeId)
        {
            $sortKey[] = $contentTypeId[0] . '-' . $contentTypeId[1];
            $contentsByType[$contentTypeId[0]][] = $contentTypeId[1];
        }

        $allContents = [];
        foreach ($contentsByType as $contentType => $ids)
        {
            $handler = $searchHandlers[$contentType]['handler'] ?? null;
            if ($handler === null)
            {
                continue;
            }
            $contents = $handler->getContent($ids, true);
            if ($contents === null)
            {
                continue;
            }
            if ($contents instanceof Entity)
            {
                $contents = new ArrayCollection([$contents]);
            }
            assert($contents instanceof AbstractCollection);
            foreach ($contents as $content)
            {
                /** @var Entity $content */
                if (method_exists($content, 'canView') && $content->canView())
                {
                    $allContents[$content->getEntityContentTypeId()] = $content;
                }
            }
        }

        return (new ArrayCollection($allContents))
            ->sortByList($sortKey)
            ->slice(0, $limit);
    }

    public function encodeForSimilarContentCache(AbstractCollection $contents): array
    {
        $blob = [];

        foreach ($contents as $content)
        {
            /** @var Entity $content */
            $blob[] = [$content->getEntityContentType(), $content->getEntityId()];
        }

        return $blob;
    }

    public function setInputRequestFromContainer(Request $request): void
    {
        // noop
    }

    public function getContainedContentTypes(): array
    {
        return [$this->contentType];
    }
}