<?php
/**
 * @noinspection PhpMissingReturnTypeInspection
 */

namespace SV\AdvancedBbCodesPack\XF\Template;

use SV\AdvancedBbCodesPack\XF\Repository\BbCode;

/**
 * @Extends \XF\Template\Templater
 *
 */
class Templater extends XFCP_Templater
{

    /**
     * @param array $controlOptions
     * @return string
     * @noinspection SpellCheckingInspection
     */
    public function formEditor(array $controlOptions)
    {
        $app = $this->app;
        $options = $app->options();
        $style = $this->getStyle() ?: $app->style();
        $bbCodeContainer = $app->bbCode();
        $container = $bbCodeContainer->container();

        $oldCustom = $container->getOriginal('custom');

        $custom = $container->offsetGet('custom');
        $custom = $this->filterCustomBbCodes($custom);
        $container->set('custom', $custom);
        $bbCodeContainer->container()->decacheFactory('rules', 'base');

        /** @var BbCode $bbCodeRepo */
        $bbCodeRepo = \XF::repository('XF:BbCode');
        $editorConfig = $bbCodeRepo->getSvEditorConfig( false);

        $dialog = $editorConfig['dialog'] ?? null;
        if (\is_array($dialog) && \count($dialog))
        {
            // array_values ensures an array not a dictionary
            $opts['dialog'] = \array_values($dialog);
        }
        if ($options->svNormalizeQuoteCharactersOnPaste ?? false)
        {
            $opts['normalizeQuoteChars'] = 1;
        }
        if (isset($custom['h1']) || isset($custom['h2']) || isset($custom['h3']) || isset($custom['h4']))
        {
            $opts['shimH1'] = 1;
        }
        if (isset($custom['hr']))
        {
            $opts['shimHr'] = 1;
        }
        if ($style->getProperty('svBbCodeEditorWordCount', false))
        {
            $opts['wordCounter'] = 1;
        }

        foreach(['abbr', 'h4'] as $tag)
        {
            if (isset($custom[$tag]))
            {
                $opts['tags'][$tag] = 1;
            }
        }
        $opts['hrStyle'] = (int)$style->getProperty('svBbcodeDefaultHrStyle');

        if (\intval($style->getProperty('svDisableQuotesInEditor')))
        {
            $opts['pasteQuotesAsBbCode'] = 1;
        }

        $disableBbCodeModeClicks = \intval($style->getProperty('svSuggestDisablingBbCodeModeClicks'));
        if ($disableBbCodeModeClicks > 0)
        {
            $opts['disableBbCodeModeClicks'] = $disableBbCodeModeClicks;
            $disableDisableBbCodeModeClicks = \intval($style->getProperty('svDisableSuggestDisablingBbCodeModeClicks'));
            if ($disableDisableBbCodeModeClicks > 0)
            {
                $opts['disableDisableBbCodeModeClicks'] = $disableDisableBbCodeModeClicks;
            }
        }

        $this->addDefaultParam('svEditorConfig', [
            'opts' => \array_filter($opts),
            'custom' => $custom,
        ]);
        try
        {
            return parent::formEditor($controlOptions);
        }
        finally
        {
            $container->set('custom', $oldCustom);
            unset($this->defaultParams['svEditorConfig']);

            // Workaround editor behaviour that a <br> should not be the first or last child of a <p> tag
            // <p><br>...</p>; editor can delete too many lines
            // <p>...<br></p>; editor can delete too few lines
            // https://xenforo.com/community/threads/editor-inserting-raw-bb-code-does-not-trim-trailing-br-tags-inside-p.196549/
            if (\XF::$versionId < 2020700)
            {
                $this->includeJs([
                    'src' => 'sv/advancedbbcode/editor/br-sanitize.js',
                    'addon' => 'SV/AdvancedBbCodesPack',
                    'min' => 1,
                ]);
            }

            // https://xenforo.com/community/threads/editor-insertspoiler-does-not-correctly-handle-formatting.210134/
            if (\XF::$versionId < 2021300)
            {
                $this->includeJs([
                    'src' => 'sv/advancedbbcode/editor/wrap-text-patch.js',
                    'addon' => 'SV/AdvancedBbCodesPack',
                    'min' => 1,
                ]);
            }

            // https://github.com/froala/wysiwyg-editor/issues/3683
            // Regression in Froala v4.0.x, appears fixed from Froala v4.0.14
            if (\XF::$versionId < 2021200)
            {
                $this->includeJs([
                    'src' => 'sv/advancedbbcode/editor/froala-patch.js',
                    'addon' => 'SV/AdvancedBbCodesPack',
                    'min' => 1,
                ]);
            }
        }
    }

    protected function filterCustomBbCodes(array $bbCodes): array
    {
        $options = \XF::options();

        // hack in Editor Manager support
        $klEditorManager = $options->offsetExists('klEMEnabledBBCodes') ? $options->klEMEnabledBBCodes : [];
        foreach ($bbCodes as $key => $config)
        {
            if (!$this->isValidBbCodeForUser($key, $config))
            {
                $klEditorManager[$key] = false;
                unset($bbCodes[$key]);
            }
        }
        if ($options->offsetExists('klEMEnabledBBCodes'))
        {
            $options->klEMEnabledBBCodes = $klEditorManager;
        }

        return $bbCodes;
    }

    /** @var array|null */
    protected $staffOnlyBbCode = null;

    /** @noinspection PhpUnusedParameterInspection */
    protected function isValidBbCodeForUser(string $tag, array $config): bool
    {
        if ($this->staffOnlyBbCode === null)
        {
            $modOnlyBbCode = \XF::options()->modOnlyBbCode ?? '';
            if (\strlen($modOnlyBbCode) !== 0 && $this->app instanceof \XF\Pub\App)
            {
                $tags = \explode(',', \str_replace("\n", ',', $modOnlyBbCode));
                $tags = \array_map('\trim', $tags);
                $tags = \array_filter($tags);
                $tags = \array_map('\strtolower', $tags);
                $this->staffOnlyBbCode = \array_fill_keys($tags, true);
            }
            else
            {
                $this->staffOnlyBbCode = [];
            }
        }

        if (isset($this->staffOnlyBbCode[$tag]))
        {
            return \XF::visitor()->is_moderator;
        }

        return true;
    }

    public function fnParseLessColor($templater, &$escape, $value)
    {
        if (!is_string($value))
        {
            return $value;
        }

        if (\preg_match('/^\s*(?:rgb|rgba)\s*\(\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*(?:,\s*(\d{1,3})\s*)?\)\s*$/i', $value, $matches))
        {
            $r = $matches[1];
            $g = $matches[2];
            $b = $matches[3];
            $a = $matches[4] ?? '';

            if ($a === '')
            {
                return \sprintf("#%02x%02x%02x", $r, $g, $b);
            }
            else
            {
                return \sprintf("#%02x%02x%02x%02x", $r, $g, $b, $a);
            }
        }

        return parent::fnParseLessColor($templater, $escape, $value);
    }

}